import { lamportToTokenAmount, USD, TurboFactory } from '@ardrive/turbo-sdk';
import { sol, createSignerFromKeypair, createGenericFileFromJson, publicKey, isKeypairSigner } from '@metaplex-foundation/umi';
import { toWeb3JsPublicKey } from '@metaplex-foundation/umi-web3js-adapters';
import { LAMPORTS_PER_SOL } from '@solana/web3.js';
import BigNumber from 'bignumber.js';
import { Buffer } from 'buffer';
import { AssetUploadFailedError, FailedToConnectToArweaveAddressError } from './errors.mjs';
import { base58 } from '../../umi-serializers-encodings/dist/esm/base58.mjs';

// Size of Arweave transaction header.
const HEADER_SIZE = 2_000;
const FREE_UPLOAD_BYTE_LIMIT = 107_520; // 105 KiB

function createArweaveUploader(context, options = {}) {
  // eslint-disable-next-line @typescript-eslint/naming-convention
  let _arweave = null;
  options = {
    solRpcUrl: context.rpc.getEndpoint(),
    ...options
  };
  const getUploadPriceFromBytes = async bytes => {
    const arweave = await getArweave();
    const wincPriceForOneSol = (await arweave.getWincForToken({
      tokenAmount: LAMPORTS_PER_SOL
    })).winc;
    const wincPriceForOneGiB = (await arweave.getUploadCosts({
      bytes: [2 ** 30]
    }))[0].winc;
    const solPriceForOneGiB = BigNumber(wincPriceForOneGiB).dividedBy(wincPriceForOneSol);
    const solPriceForOneByte = solPriceForOneGiB.dividedBy(2 ** 30);
    const solPriceForGivenBytes = solPriceForOneByte.multipliedBy(bytes);
    const wincPriceForGivenBytes = BigNumber(wincPriceForOneGiB).multipliedBy(bytes / 2 ** 30).integerValue(BigNumber.ROUND_CEIL);
    const solAmount = sol(solPriceForGivenBytes.multipliedBy(options.priceMultiplier ?? 1.1).toNumber());
    return {
      solAmount,
      wincAmount: wincPriceForGivenBytes
    };
  };
  const getBytesFromFiles = files => files.reduce((sum, file) => sum + HEADER_SIZE + file.buffer.byteLength, 0);
  const getUploadSolAndWincPrice = async files => {
    const bytes = getBytesFromFiles(files);
    if (bytes <= FREE_UPLOAD_BYTE_LIMIT) {
      return {
        solAmount: sol(0),
        wincAmount: new BigNumber(0)
      };
    }
    return getUploadPriceFromBytes(bytes);
  };
  const getUploadPrice = async files => {
    const {
      solAmount
    } = await getUploadSolAndWincPrice(files);
    return solAmount;
  };
  const upload = async files => {
    const arweave = await getArweave();
    const {
      solAmount,
      wincAmount
    } = await getUploadSolAndWincPrice(files);
    await fund(solAmount);
    const ephemeralKey = context.eddsa.generateKeypair();
    const ephemeralArweave = await initArweave(createSignerFromKeypair(context, ephemeralKey));
    if (wincAmount.isGreaterThan(0)) {
      // Share credits to ephemeral key
      await arweave.shareCredits({
        approvedAddress: await ephemeralArweave.signer.getNativeAddress(),
        approvedWincAmount: wincAmount,
        // TODO: Could make configurable to prevent upload timeouts
        expiresBySeconds: 60 * 20 // Provides 20 minutes to upload, expiring any un-used credits back to the payer.
      });
    }

    const paidBy = await arweave.signer.getNativeAddress();
    const promises = files.map(async file => {
      const buffer = Buffer.from(file.buffer);
      let dataItemId;
      try {
        // Send uploads with ephemeral key to avoid signing each upload with the payer's key.
        const {
          id
        } = await ephemeralArweave.uploadFile({
          fileStreamFactory: () => buffer,
          fileSizeFactory: () => buffer.byteLength,
          dataItemOpts: {
            tags: getGenericFileTagsWithContentType(file),
            paidBy
          }
        });
        dataItemId = id;
      } catch (error) {
        throw new AssetUploadFailedError(error);
      }
      const gateway = context.rpc.getCluster() === 'devnet' ? 'https://turbo.ardrive.dev/raw' : 'https://arweave.net';
      return `${gateway}/${dataItemId}`;
    });
    return Promise.all(promises);
  };
  const uploadJson = async json => {
    const file = createGenericFileFromJson(json);
    const uris = await upload([file]);
    return uris[0];
  };
  const getTurboStorageCreditBalance = async () => {
    const {
      winc
    } = await (await getArweave()).getBalance();
    return winc;
  };
  const getBalance = async () => {
    const usersWincBalance = await getTurboStorageCreditBalance();
    const wincForOneSol = (await (await getArweave()).getWincForToken({
      tokenAmount: LAMPORTS_PER_SOL
    })).winc;
    const solEquivalent = BigNumber(usersWincBalance).dividedBy(wincForOneSol);
    return sol(solEquivalent.toNumber());
  };
  const fund = async (amount, skipBalanceCheck = false) => {
    let toFundLamports = amountToBigNumber(amount);
    if (!skipBalanceCheck) {
      const lamportEquivalentBalance = amountToBigNumber(await getBalance());
      toFundLamports = toFundLamports.isGreaterThan(lamportEquivalentBalance) ? toFundLamports.minus(lamportEquivalentBalance) : new BigNumber(0);
    }
    if (toFundLamports.isLessThanOrEqualTo(0)) {
      return;
    }
    await (await getArweave()).topUpWithTokens({
      tokenAmount: lamportToTokenAmount(toFundLamports)
    });
  };
  const getStripeCheckoutSession = async amount => {
    const arweave = await getArweave();
    const usdAmount = amountToBigNumber(amount).shiftedBy(-2).toNumber();
    const {
      winc,
      url
    } = await arweave.createCheckoutSession({
      amount: USD(usdAmount),
      owner: toWeb3JsPublicKey(context.payer.publicKey).toBase58()
    });
    if (!url) {
      throw new Error('Failed to create Stripe Checkout Session');
    }
    return {
      turboStorageCredits: BigNumber(winc).exponentiatedBy(12),
      checkoutSessionUrl: url
    };
  };
  const getArweave = async () => {
    const oldPayer = await _arweave?.signer.getPublicKey();
    const newPayer = options.payer ?? context.payer;
    if (oldPayer && publicKey(new Uint8Array(oldPayer)) !== newPayer.publicKey) {
      _arweave = null;
    }
    if (!_arweave) {
      _arweave = await initArweave();
    }
    return _arweave;
  };
  const defaultAddresses = context.rpc.getCluster() === 'devnet' ? {
    uploadServiceUrl: 'https://upload.ardrive.dev',
    paymentServiceUrl: 'https://payment.ardrive.dev'
  } : {
    uploadServiceUrl: 'https://upload.ardrive.io',
    paymentServiceUrl: 'https://payment.ardrive.io'
  };
  const initArweave = async (payer = options.payer ?? context.payer) => {
    let arweave;
    if (isKeypairSigner(payer)) {
      arweave = TurboFactory.authenticated({
        token: 'solana',
        privateKey: base58.deserialize(payer.secretKey)[0],
        gatewayUrl: options.solRpcUrl,
        uploadServiceConfig: {
          url: options.uploadServiceUrl ?? defaultAddresses.uploadServiceUrl
        },
        paymentServiceConfig: {
          url: options.paymentServiceUrl ?? defaultAddresses.paymentServiceUrl
        }
      });
    } else {
      arweave = await initWebArweave(payer, options);
    }
    try {
      // Check for valid Arweave node.
      await arweave.getTurboCryptoWallets();
    } catch (error) {
      throw new FailedToConnectToArweaveAddressError(options.uploadServiceUrl ?? '', options.paymentServiceUrl ?? '', error);
    }
    return arweave;
  };
  const initWebArweave = async (payer, options) => {
    const wallet = {
      publicKey: toWeb3JsPublicKey(payer.publicKey),
      signMessage: payer.signMessage
    };
    return TurboFactory.authenticated({
      token: 'solana',
      walletAdapter: wallet,
      gatewayUrl: options.solRpcUrl,
      uploadServiceConfig: {
        url: options.uploadServiceUrl ?? defaultAddresses.uploadServiceUrl
      },
      paymentServiceConfig: {
        url: options.paymentServiceUrl ?? defaultAddresses.paymentServiceUrl
      }
    });
  };
  return {
    getUploadPriceFromBytes,
    getUploadPrice,
    upload,
    uploadJson,
    getBalance,
    getTurboStorageCreditBalance,
    fund,
    getStripeCheckoutSession,
    arweave: getArweave
  };
}
const isArweaveUploader = uploader => 'arweave' in uploader && 'getBalance' in uploader && 'fund' in uploader;
const amountToBigNumber = amount => new BigNumber(amount.basisPoints.toString());
const getGenericFileTagsWithContentType = file => {
  if (!file.contentType) {
    return file.tags;
  }
  return [{
    name: 'Content-Type',
    value: file.contentType
  }, ...file.tags];
};

export { createArweaveUploader, isArweaveUploader };
//# sourceMappingURL=createArweaveUploader.mjs.map
